#include <cairo/cairo.h>
#include <string.h>
#include <assert.h>
#include <limits.h>
#include "cairo-widget.h"
/*
#include <cairo.h>

int main()
{
    cairo_surface_t *s = cairo_image_surface_create(CAIRO_FORMAT_RGB24, 150, 50);
    cairo_t *cr = cairo_create(s);

    cairo_set_source_rgb(cr, 1, 0, 0);
    cairo_paint(cr);

    cairo_rectangle(cr, 0, 0, 100, 100);
    cairo_clip(cr);
    cairo_move_to(cr, 50, 25);
    cairo_set_source_rgb(cr, 0, 0, 0);
    cairo_show_text(cr, "pretend that this string is > 100px wide");

    cairo_destroy(cr);
    cairo_surface_write_to_png(s, "out.png");
    cairo_surface_destroy(s);

    return 0;
}
*/
enum
{
    CW_NO_PROP,
    CW_EXTENT_DEF,
    CW_EXTENT,
    CW_PARENT,
    CW_VISIBILITY,
    CW_HALIGN,
    CW_VALIGN,
    CW_HSPACE,
    CW_VSPACE,
    CW_PADDING,
    CW_PADDING_LEFT,
    CW_PADDING_TOP,
    CW_PADDING_BOTTOM,
    CW_PADDING_RIGHT,
    CW_MARGIN_LEFT,
    CW_MARGIN_RIGHT,
    CW_MARGIN_TOP,
    CW_MARGIN_BOTTOM,
    CW_MARGIN,
    CW_FILL,
    CW_STROKE,
    CW_N_PROPERTIES
};
static GParamSpec *g_cw_prop[CW_N_PROPERTIES] = {0};

typedef struct _DiyaCairoWidgetPrivate
{
    DiyaObject * parent_object;
    diya_rect_t extent;
    diya_rect_t default_extent;
    DiyaCairoWidget * parent_widget;

    gchar string[64];
    
    guint16 visibility;
    guint16 halign;
    guint16 valign;
    guint16 hspace;
    guint16 vspace;

    diya_offset_t margin;
    diya_offset_t padding;

    diya_color_t fill;
    
    diya_stroke_t stroke;

    gboolean damaged;
} DiyaCairoWidgetPrivate;

G_DEFINE_TYPE_WITH_PRIVATE(DiyaCairoWidget, diya_cairo_widget, DIYA_TYPE_OBJECT);


static void diya_cairo_widget_dispose(GObject* object)
{
    g_debug("diya_cairo_widget_dispose: %s", diya_object_to_string(object));
    G_OBJECT_CLASS(diya_cairo_widget_parent_class)->dispose(object);
}

static void diya_cairo_widget_set_property(GObject *object, guint property_id, const GValue *value, GParamSpec *pspec)
{
    DiyaCairoWidget * self = DIYA_CAIRO_WIDGET(object);
    DiyaCairoWidgetPrivate* priv = diya_cairo_widget_get_instance_private(self);
    gpointer ptr;
    switch (property_id)
    {
    case CW_PARENT:
        priv->parent_widget = g_value_get_pointer(value);
        assert(DIYA_IS_CAIRO_WIDGET(priv->parent_widget));
        break;
    case CW_EXTENT_DEF:
        ptr = g_value_get_pointer(value);
        assert(ptr);
        priv->default_extent = *(diya_rect_t*)ptr;
        break;
    case CW_EXTENT:
        break;
    case CW_VISIBILITY:
        priv->visibility = (guint16) g_value_get_uint(value);
        break;
    case CW_HALIGN:
        priv->halign = (guint16)g_value_get_uint(value);
        break;
    case CW_VALIGN:
        priv->valign = (guint16)g_value_get_uint(value);
        break;
    case CW_HSPACE:
        priv->hspace = (guint16)g_value_get_uint(value);
        break;
    case CW_VSPACE:
        priv->vspace = (guint16)g_value_get_uint(value);
        break;
    case CW_PADDING:
        ptr = g_value_get_pointer(value);
        assert(ptr);
        priv->padding = *(diya_offset_t*)ptr;
        break;
    case CW_PADDING_LEFT:
        priv->padding.left = g_value_get_int(value);
        break;
    case CW_PADDING_RIGHT:
        priv->padding.right = g_value_get_int(value);
        break;
    case CW_PADDING_BOTTOM:
        priv->padding.bottom = g_value_get_int(value);
        break;
    case CW_PADDING_TOP:
        priv->padding.top = g_value_get_int(value);
        break;
    case CW_MARGIN:
        ptr = g_value_get_pointer(value);
        assert(ptr);
        priv->margin = *(diya_offset_t*)ptr;
        break;
    case CW_MARGIN_LEFT:
        priv->margin.left = g_value_get_int(value);
        break;
    case CW_MARGIN_RIGHT:
        priv->margin.right = g_value_get_int(value);
        break;
    case CW_MARGIN_BOTTOM:
        priv->margin.bottom = g_value_get_int(value);
        break;
    case CW_MARGIN_TOP:
        priv->margin.top = g_value_get_int(value);
        break;
    case CW_FILL:
        ptr = g_value_get_pointer(value);
        assert(ptr);
        priv->fill = *(diya_color_t*)ptr;
        break;
    case CW_STROKE:
        ptr = g_value_get_pointer(value);
        assert(ptr);
        priv->stroke = *(diya_stroke_t*) ptr;
        break;
    default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
        break;
    }
}

static void diya_cairo_widget_get_property(GObject *object, guint property_id, GValue *value, GParamSpec *pspec)
{
    DiyaCairoWidget * self = DIYA_CAIRO_WIDGET(object);
    DiyaCairoWidgetPrivate* priv = diya_cairo_widget_get_instance_private(self);
    switch (property_id)
    {
    case CW_PARENT:
        g_value_set_pointer(value, priv->parent_widget);
        break;
    case CW_EXTENT_DEF:
        g_value_set_pointer(value, &priv->default_extent);
        break;
    case CW_EXTENT:
        g_value_set_pointer(value, &priv->extent);
        break;
    case CW_VISIBILITY:
        g_value_set_uint(value, priv->visibility);
        break;
    case CW_HALIGN:
        g_value_set_uint(value, priv->halign);
        break;
    case CW_VALIGN:
        g_value_set_uint(value, priv->valign);
        break;
    case CW_HSPACE:
        g_value_set_uint(value, priv->hspace);
        break;
    case CW_VSPACE:
        g_value_set_uint(value, priv->vspace);
        break;
    case CW_PADDING:
        g_value_set_pointer(value, &priv->padding);
        break;
    case CW_MARGIN:
        g_value_set_pointer(value, &priv->margin);
        break;
    case CW_FILL:
        g_value_set_pointer(value, &priv->fill);
        break;
    case CW_STROKE:
        g_value_set_pointer(value, &priv->stroke);
        break;
    default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
        break;
    }
}

static void diya_cairo_widget_init(DiyaCairoWidget *self)
{
    DiyaCairoWidgetPrivate* priv = diya_cairo_widget_get_instance_private(self);
    memset(&priv->extent,0,sizeof(priv->extent));
    memset(&priv->extent,0,sizeof(priv->default_extent));
    priv->parent_widget = NULL;
    priv->string[0] = '\0';

    priv->visibility = DIYA_CAIRO_WIDGET_VISIBLE_ON;

    priv->hspace = DIYA_CAIRO_WIDGET_FILL_EXTEND;
    priv->vspace = DIYA_CAIRO_WIDGET_FILL_EXTEND;

    priv->halign = DIYA_CAIRO_WIDGET_ALIGN_MIDDLE;
    priv->valign = DIYA_CAIRO_WIDGET_ALIGN_MIDDLE;

    memset(&priv->padding,0,sizeof(priv->padding));
    memset(&priv->margin,0,sizeof(priv->margin));
    memset(&priv->fill, 0, sizeof(priv->fill));

    priv->damaged = TRUE;
}

static const gchar *diya_cairo_widget_to_string(DiyaObject *object)
{
    DiyaCairoWidget * self = DIYA_CAIRO_WIDGET(object);
    DiyaCairoWidgetClass *class = DIYA_CAIRO_WIDGET_GET_CLASS(object);
    DiyaCairoWidgetPrivate* priv = diya_cairo_widget_get_instance_private(self);

    g_snprintf(priv->string,sizeof(priv->string),
        "%s (%dx%d) at %d@%d",
        class->name,
        priv->extent.x, priv->extent.y,
        priv->extent.width, priv->extent.height
    );
    return priv->string;
}

static diya_rect_t diya_cairo_widget_get_view_port(gpointer object)
{
    DiyaCairoWidget * self = DIYA_CAIRO_WIDGET(object);
    DiyaCairoWidgetPrivate* priv = diya_cairo_widget_get_instance_private(self);

    diya_point_t gpos = diya_cairo_widget_get_global_position(object);
    diya_rect_t view = {gpos.x, gpos.y, priv->extent.width, priv->extent.height};
    if(priv->parent_widget)
    {
        diya_rect_t parent_view = diya_cairo_widget_get_view_port(priv->parent_widget);

        if(!diya_rect_overlap(&view, &parent_view))
        {
            view.x = view.y = view.width = view.height = 0;
            return view;
        }

        view.width = MIN(view.x + view.width, parent_view.x + parent_view.width);
        view.height = MIN(view.y + view.height, parent_view.y + parent_view.height);

        view.x = MAX(view.x, parent_view.x);
        view.y = MAX(view.y, parent_view.y);
        
        view.width -= view.x;
        view.height -= view.y;

    }
    return view;
}

static void diya_cairo_widget_draw(gpointer object, cairo_t* context)
{
    (void) context;

    DiyaCairoWidget * self = DIYA_CAIRO_WIDGET(object);
    DiyaCairoWidgetPrivate* priv = diya_cairo_widget_get_instance_private(self);

    if(!priv->damaged)
    {
        return;
    }
    if(priv->visibility != DIYA_CAIRO_WIDGET_VISIBLE_ON)
    {
        return;
    }
    diya_rect_t view_port = diya_cairo_widget_get_view_port(object);
    if(diya_rect_is_empty(&view_port))
    {
        return;
    }
    cairo_save(context);
    cairo_rectangle(context, view_port.x, view_port.y, view_port.width, view_port.height);
    cairo_clip(context);
    diya_point_t gpos = diya_cairo_widget_get_global_position(object);
    cairo_move_to(context, (double)gpos.x, (double)gpos.y);
    if(priv->fill.a > 0)
    {
        cairo_set_source_rgba(context, priv->fill.r, priv->fill.g, priv->fill.b, priv->fill.a);
        cairo_fill(context);
    }
    if(priv->stroke.size > 0 && priv->stroke.color.a != 0)
    {
        cairo_set_line_width(context,priv->stroke.size);
        cairo_stroke(context);
    }
    cairo_restore(context);
    priv->damaged = FALSE;
}

static void diya_cairo_widget_content_size(gpointer object, guint* width, guint* height)
{
    DiyaCairoWidget * self = DIYA_CAIRO_WIDGET(object);
    DiyaCairoWidgetPrivate* priv = diya_cairo_widget_get_instance_private(self);
    *width = priv->extent.width;
    *height = priv->extent.height;
    // take into account margin
}
static void diya_cairo_widget_resize(gpointer object)
{
    // calculate effective size
    DiyaCairoWidget * self = DIYA_CAIRO_WIDGET(object);
    DiyaCairoWidgetPrivate* priv = diya_cairo_widget_get_instance_private(self);
    if(priv->visibility == DIYA_CAIRO_WIDGET_VISIBLE_HIDDEN)
    {
        /**
         * Item is hidden, set extent to empty
        */
       if(!diya_rect_is_empty(&priv->extent))
       {
            priv->damaged = TRUE;
       }
        priv->extent.width = 0;
        priv->extent.height = 0;
        priv->extent.x = 0;
        priv->extent.y = 0;
        return;
    }
    diya_rect_t space = {
        priv->margin.left,
        priv->margin.top,
        priv->extent.width - priv->margin.left - priv->margin.right,
        priv->extent.height - priv->margin.top - priv->margin.bottom};
    if(priv->parent_widget)
    {
        space = DIYA_CAIRO_WIDGET_GET_CLASS(self)->alloc_space(priv->parent_object, object);
    }
    guint width,height,cwidth = 0,cheight = 0;
    if(priv->hspace == DIYA_CAIRO_WIDGET_FILL_CONTENT ||
        priv->vspace == DIYA_CAIRO_WIDGET_FILL_CONTENT)
        {
            DIYA_CAIRO_WIDGET_GET_CLASS(self)->content_size(object, &cwidth, &cheight);
        }
    switch(priv->hspace)
    {
        case DIYA_CAIRO_WIDGET_FILL_NONE:
            width = priv->default_extent.width;
            break;
        case DIYA_CAIRO_WIDGET_FILL_EXTEND:
            width = space.width;
            break;
        case DIYA_CAIRO_WIDGET_FILL_CONTENT:
            width = cwidth;
            break;
        default:
            width = priv->extent.width;
            break;
    }

    switch(priv->vspace)
    {
        case DIYA_CAIRO_WIDGET_FILL_NONE:
            height = priv->default_extent.height;
            break;
        case DIYA_CAIRO_WIDGET_FILL_EXTEND:
            height = space.height;
            break;
        case DIYA_CAIRO_WIDGET_FILL_CONTENT:
            height = cheight;
            break;
        default:
            height = priv->extent.height;
            break;
    }

    if(width != priv->extent.width || height != priv->extent.height)
    {
        priv->damaged = TRUE;
    }
    priv->extent.width = width;
    priv->extent.height = height;
    /*if(space.x != priv->extent.x || space.y != priv->extent.y)
    {
        priv->damaged = TRUE;
    }*/
    priv->extent.x = space.x;
    priv->extent.y = space.y;
}
static void diya_cairo_widget_align(gpointer object)
{
    guint cwidth, cheight;
    // calculate widget position
    DiyaCairoWidget * self = DIYA_CAIRO_WIDGET(object);
    DiyaCairoWidgetPrivate* priv = diya_cairo_widget_get_instance_private(self);
    DIYA_CAIRO_WIDGET_GET_CLASS(self)->content_size(object, &cwidth, &cheight);
    if(diya_rect_is_empty(&priv->extent))
    {
        return;
    }
    int x = priv->extent.x, y = priv->extent.y;
    switch(priv->halign)
    {
        case DIYA_CAIRO_WIDGET_ALIGN_MIDDLE:
            x += (priv->extent.width - cwidth ) / 2;
            break;
        case DIYA_CAIRO_WIDGET_ALIGN_END:
            x += priv->extent.width - cwidth;
            break;
        default:
            break;
    }

    switch(priv->valign)
    {
        case DIYA_CAIRO_WIDGET_ALIGN_MIDDLE:
            y += (priv->extent.height - cheight ) / 2;
            break;
        case DIYA_CAIRO_WIDGET_ALIGN_END:
            y += priv->extent.height - cheight;
            break;
        default:
            break;
    }
    if(x != priv->extent.x || y != priv->extent.y)
    {
        priv->damaged = TRUE;
    }
    priv->extent.x = x;
    priv->extent.y = y;
}
static void diya_cairo_widget_update(gpointer object)
{
    
    DIYA_CAIRO_WIDGET_GET_CLASS(DIYA_CAIRO_WIDGET(object))->resize(object);
    DIYA_CAIRO_WIDGET_GET_CLASS(DIYA_CAIRO_WIDGET(object))->align(object);
}

static diya_rect_t diya_cairo_widget_alloc_space(gpointer object, gpointer child)
{
    assert(object);
    DiyaCairoWidget * self = DIYA_CAIRO_WIDGET(object);
    DiyaCairoWidgetPrivate* priv = diya_cairo_widget_get_instance_private(self);
    diya_rect_t space = {0,0,priv->extent.width, priv->extent.height};
    diya_offset_t * margin = NULL;
    g_object_get(child,"margin", &margin, NULL);

    space.x += priv->padding.left + margin->left;
    space.y += priv->padding.top + margin->top;
    space.width -= (priv->padding.right + priv->padding.left + margin->right + margin->left);
    space.height -= (priv->padding.bottom + priv->padding.top + margin->bottom + margin->top);

    return space;
}

static void diya_cairo_widget_class_init(DiyaCairoWidgetClass *class)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS(class);
    DiyaObjectClass *base_class = DIYA_OBJECT_CLASS(class);
    DIYA_CAIRO_WIDGET_SET_NAME(class, "DiyaCairoWidget");
    gobject_class->dispose = diya_cairo_widget_dispose;
    gobject_class->set_property = diya_cairo_widget_set_property;
    gobject_class->get_property = diya_cairo_widget_get_property;
    class->draw = diya_cairo_widget_draw;
    class->update = diya_cairo_widget_update;
    class->alloc_space = diya_cairo_widget_alloc_space;
    class->resize = diya_cairo_widget_resize;
    class->align = diya_cairo_widget_align;
    class->content_size = diya_cairo_widget_content_size;

    g_cw_prop[CW_EXTENT_DEF] = g_param_spec_pointer("default-extent", NULL, "Widget actual extent relative to its parent", G_PARAM_READWRITE); //
    g_cw_prop[CW_EXTENT] = g_param_spec_pointer("extent", NULL, "Widget default extent", G_PARAM_READABLE); 
    g_cw_prop[CW_PARENT] = g_param_spec_pointer("parent", NULL, "Parent widget", G_PARAM_READWRITE); // 
    g_cw_prop[CW_FILL] = g_param_spec_pointer("fill", NULL, "Fill color", G_PARAM_READWRITE);
    g_cw_prop[CW_STROKE] = g_param_spec_pointer("stroke", NULL, "Border size and color", G_PARAM_READWRITE);
    g_cw_prop[CW_VISIBILITY] = g_param_spec_uint(
            "visibility", NULL, "Widget visibility",
            DIYA_CAIRO_WIDGET_VISIBLE_ON,
            DIYA_CAIRO_WIDGET_VISIBLE_MAX,
            DIYA_CAIRO_WIDGET_VISIBLE_ON,
            G_PARAM_READWRITE);

    g_cw_prop[CW_HALIGN] = g_param_spec_uint(
            "halign", NULL, "Widget halign",
            DIYA_CAIRO_WIDGET_ALIGN_START,
            DIYA_CAIRO_WIDGET_ALIGN_MAX,
            DIYA_CAIRO_WIDGET_ALIGN_MIDDLE,
            G_PARAM_READWRITE);
    
    g_cw_prop[CW_VALIGN] = g_param_spec_uint(
            "valign", NULL, "Widget valign",
            DIYA_CAIRO_WIDGET_ALIGN_START,
            DIYA_CAIRO_WIDGET_ALIGN_MAX,
            DIYA_CAIRO_WIDGET_ALIGN_MIDDLE,
            G_PARAM_READWRITE);
    
    g_cw_prop[CW_HSPACE] = g_param_spec_uint(
            "hspace", NULL, "Widget hspace",
            DIYA_CAIRO_WIDGET_FILL_NONE,
            DIYA_CAIRO_WIDGET_FILL_MAX,
            DIYA_CAIRO_WIDGET_FILL_EXTEND,
            G_PARAM_READWRITE);
    
    g_cw_prop[CW_VSPACE] = g_param_spec_uint(
            "vpsace", NULL, "Widget vspace",
            DIYA_CAIRO_WIDGET_FILL_NONE,
            DIYA_CAIRO_WIDGET_FILL_MAX,
            DIYA_CAIRO_WIDGET_FILL_EXTEND,
            G_PARAM_READWRITE);

    g_cw_prop[CW_PADDING] = g_param_spec_pointer("padding", NULL, "Widget padding", G_PARAM_READWRITE); 
    g_cw_prop[CW_MARGIN] = g_param_spec_pointer("margin", NULL, "Widget margin", G_PARAM_READWRITE);

    g_cw_prop[CW_PADDING_LEFT] = g_param_spec_int(
            "padding-left", NULL, "Widget padding-left",
            G_MININT,
            G_MAXINT,
            0,
            G_PARAM_WRITABLE);
    g_cw_prop[CW_PADDING_RIGHT] = g_param_spec_int(
            "padding-right", NULL, "Widget padding-right",
            INT_MIN,
            INT_MAX,
            0,
            G_PARAM_WRITABLE);
    g_cw_prop[CW_PADDING_TOP] = g_param_spec_int(
            "padding-top", NULL, "Widget padding-top",
            INT_MIN,
            INT_MAX,
            0,
            G_PARAM_WRITABLE);
    g_cw_prop[CW_PADDING_BOTTOM] = g_param_spec_int(
            "padding-bottom", NULL, "Widget padding-bottom",
            INT_MIN,
            INT_MAX,
            0,
            G_PARAM_WRITABLE);
    
    g_cw_prop[CW_MARGIN_LEFT] = g_param_spec_int(
            "margin-left", NULL, "Widget margin-left",
            INT_MIN,
            INT_MAX,
            0,
            G_PARAM_WRITABLE);
    g_cw_prop[CW_MARGIN_RIGHT] = g_param_spec_int(
            "margin-right", NULL, "Widget margin-right",
            INT_MIN,
            INT_MAX,
            0,
            G_PARAM_WRITABLE);
    g_cw_prop[CW_MARGIN_TOP] = g_param_spec_int(
            "margin-top", NULL, "Widget margin-top",
            INT_MIN,
            INT_MAX,
            0,
            G_PARAM_WRITABLE);
    g_cw_prop[CW_MARGIN_BOTTOM] = g_param_spec_int(
            "margin-bottom", NULL, "Widget margin-bottom",
            INT_MIN,
            INT_MAX,
            0,
            G_PARAM_WRITABLE);

    base_class->to_string = diya_cairo_widget_to_string;

    g_object_class_install_properties (gobject_class, CW_N_PROPERTIES, g_cw_prop);
}

diya_point_t diya_cairo_widget_get_position(gpointer object)
{
    diya_point_t p = {0};
    assert(DIYA_IS_CAIRO_WIDGET(object));
    DiyaCairoWidget * self = DIYA_CAIRO_WIDGET(object);
    DiyaCairoWidgetPrivate* priv = diya_cairo_widget_get_instance_private(self);
    p.x = priv->extent.x;
    p.y = priv->extent.y;
    return p;
}
diya_point_t diya_cairo_widget_get_global_position(gpointer object)
{
    assert(DIYA_IS_CAIRO_WIDGET(object));
    DiyaCairoWidget * self = DIYA_CAIRO_WIDGET(object);
    DiyaCairoWidgetPrivate* priv = diya_cairo_widget_get_instance_private(self);
    diya_point_t p = {priv->extent.x, priv->extent.y};
    if(priv->parent_widget)
    {
        
        diya_point_t parent_pos = diya_cairo_widget_get_global_position(priv->parent_widget);
        p.x += parent_pos.x;
        p.y += parent_pos.y;
    }
    return p;
}
DiyaCairoWidget* diya_cairo_widget_get_parent(gpointer object)
{
    assert(DIYA_IS_CAIRO_WIDGET(object));
    DiyaCairoWidget * self = DIYA_CAIRO_WIDGET(object);
    DiyaCairoWidgetPrivate* priv = diya_cairo_widget_get_instance_private(self);
    return priv->parent_widget;
}
diya_rect_t diya_cairo_widget_get_extent(gpointer object)
{
    assert(DIYA_IS_CAIRO_WIDGET(object));
    DiyaCairoWidget * self = DIYA_CAIRO_WIDGET(object);
    DiyaCairoWidgetPrivate* priv = diya_cairo_widget_get_instance_private(self);

    return priv->extent;
}
gboolean diya_cairo_widget_is_visible(gpointer object)
{
    assert(DIYA_IS_CAIRO_WIDGET(object));
    DiyaCairoWidget * self = DIYA_CAIRO_WIDGET(object);
    DiyaCairoWidgetPrivate* priv = diya_cairo_widget_get_instance_private(self);
    return priv->visibility == DIYA_CAIRO_WIDGET_VISIBLE_ON;
}
gboolean diya_cairo_widget_is_hidden(gpointer object)
{
    assert(DIYA_IS_CAIRO_WIDGET(object));
    DiyaCairoWidget * self = DIYA_CAIRO_WIDGET(object);
    DiyaCairoWidgetPrivate* priv = diya_cairo_widget_get_instance_private(self);
    return priv->visibility == DIYA_CAIRO_WIDGET_VISIBLE_HIDDEN;
}

void diya_cairo_widget_set_size(gpointer object, guint width, guint height)
{
    assert(DIYA_IS_CAIRO_WIDGET(object));
    DiyaCairoWidget * self = DIYA_CAIRO_WIDGET(object);
    DiyaCairoWidgetPrivate* priv = diya_cairo_widget_get_instance_private(self);

    priv->default_extent.width = width;
    priv->default_extent.height = height;

}
void diya_cairo_widget_set_position(gpointer object, gint x, gint y)
{
    assert(DIYA_IS_CAIRO_WIDGET(object));
    DiyaCairoWidget * self = DIYA_CAIRO_WIDGET(object);
    DiyaCairoWidgetPrivate* priv = diya_cairo_widget_get_instance_private(self);
    
    priv->default_extent.x = x;
    priv->default_extent.y = y;
    
}

void diya_cairo_widget_fill(gpointer object, diya_color_t * color)
{
    assert(DIYA_IS_CAIRO_WIDGET(object));
    DiyaCairoWidget * self = DIYA_CAIRO_WIDGET(object);
    DiyaCairoWidgetPrivate* priv = diya_cairo_widget_get_instance_private(self);
    priv->fill = *color;
}
void diya_cairo_widget_set_stroke_size(gpointer object, gdouble size)
{
    assert(DIYA_IS_CAIRO_WIDGET(object));
    DiyaCairoWidget * self = DIYA_CAIRO_WIDGET(object);
    DiyaCairoWidgetPrivate* priv = diya_cairo_widget_get_instance_private(self);

    priv->stroke.size = size;
}
void diya_cairo_widget_set_stroke_color(gpointer object, diya_color_t * color)
{
    assert(DIYA_IS_CAIRO_WIDGET(object));
    DiyaCairoWidget * self = DIYA_CAIRO_WIDGET(object);
    DiyaCairoWidgetPrivate* priv = diya_cairo_widget_get_instance_private(self);

    priv->stroke.color = *color;
}

gboolean diya_cairo_widget_is_damaged(gpointer object)
{
    assert(DIYA_IS_CAIRO_WIDGET(object));
    DiyaCairoWidget * self = DIYA_CAIRO_WIDGET(object);
    DiyaCairoWidgetPrivate* priv = diya_cairo_widget_get_instance_private(self);

    return priv->damaged;
}

void diya_cairo_widget_damage(gpointer object)
{
    assert(DIYA_IS_CAIRO_WIDGET(object));
    DiyaCairoWidget * self = DIYA_CAIRO_WIDGET(object);
    DiyaCairoWidgetPrivate* priv = diya_cairo_widget_get_instance_private(self);

    priv->damaged = TRUE;
}

/*Container API*/

typedef struct _DiyaCairoContainerPrivate
{
    DiyaCairoWidget * parent_object;
    GList* children;

} DiyaCairoContainerPrivate;

G_DEFINE_TYPE_WITH_PRIVATE(DiyaCairoContainer, diya_cairo_container, DIYA_TYPE_CAIRO_WIDGET)

static void diya_cairo_container_dispose(GObject* object)
{
    DiyaCairoContainer * self = DIYA_CAIRO_CONTAINER(object);
    DiyaCairoContainerPrivate* priv = diya_cairo_container_get_instance_private(self);
    g_debug("diya_cairo_container_dispose: %s", diya_object_to_string(object));
    g_list_free_full(priv->children, g_object_unref);
    G_OBJECT_CLASS(diya_cairo_widget_parent_class)->dispose(object);
}

static void diya_cairo_container_init(DiyaCairoContainer *self)
{
    DiyaCairoContainerPrivate* priv = diya_cairo_container_get_instance_private(self);
    priv->children = NULL;
}
static void diya_cairo_container_class_init(DiyaCairoContainerClass *class)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS(class);
    //DiyaObjectClass *base_class = DIYA_OBJECT_CLASS(class);
    DIYA_CAIRO_WIDGET_SET_NAME(class, "DiyaCairoContainer");
    gobject_class->dispose = diya_cairo_container_dispose;
}

void diya_cairo_container_add(gpointer object, gpointer child)
{
    assert(DIYA_IS_CAIRO_WIDGET(child));
    assert(DIYA_IS_CAIRO_CONTAINER(object));
    DiyaCairoContainerPrivate* priv = diya_cairo_container_get_instance_private(DIYA_CAIRO_CONTAINER(object));
    g_object_set(child,"parent", object, NULL);
    if(g_list_find (priv->children,child))
    {
        return;
    }
    priv->children = g_list_append(priv->children, child);
}

void diya_offset_normalize(diya_offset_t * offset)
{
    gint tmp;
    if(offset->left > offset->right)
    {
        tmp = offset->left;
        offset->left = offset->right;
        offset->right = tmp;
    }
    if(offset->top > offset->bottom)
    {
        tmp = offset->top;
        offset->top = offset->bottom;
        offset->bottom = tmp;
    }
}