#ifndef CAIRO_WIDGET_H
#define CAIRO_WIDGET_H


#include <cairo/cairo.h>
#include "../base.h"

// basic type

typedef struct
{
    gint x;
    gint y;
    guint width;
    guint height;
} diya_rect_t;

typedef struct
{
    gint x;
    gint y;
}
diya_point_t;

typedef struct
{
    gint left;
    gint top;
    gint bottom;
    gint right;
}
diya_offset_t;

typedef struct
{
    gdouble r;
    gdouble g;
    gdouble b;
    gdouble a;
}
diya_color_t;

typedef struct
{
    gdouble size;
    diya_color_t color;
} diya_stroke_t;

#define BETWEEN(value,min,max) ((value > min) && (value < max))

#define diya_rect_is_empty(rect) (((rect)->width <= 0) && ((rect)->height <= 0))
#define diya_rect_to_offset(rect) ({(rect)->x,(rect)->y,(rect)->y+(rect)->height, (rect)->x + (rect)->width})
#define diya_rect_overlap(a,b) ( \
    (BETWEEN((a)->x, (b)->x, (b)->x + (gint)(b)->width) || BETWEEN((b)->x, (a)->x, (a)->x + (gint)(a)->width)) && \
    (BETWEEN((a)->y, (b)->y, (b)->y + (gint)(b)->height) || BETWEEN((b)->y, (a)->y, (a)->y + (gint)(a)->height)))

void diya_offset_normalize(diya_offset_t * offset);

#define DIYA_CAIRO_WIDGET_SET_NAME(c, n) (DIYA_CAIRO_WIDGET_CLASS(c)->name = n)
#define DIYA_TYPE_CAIRO_WIDGET  (diya_cairo_widget_get_type ())
G_DECLARE_DERIVABLE_TYPE (DiyaCairoWidget, diya_cairo_widget, DIYA, CAIRO_WIDGET, DiyaObject)

enum
{
    /*Align enum*/
    DIYA_CAIRO_WIDGET_ALIGN_START,
    DIYA_CAIRO_WIDGET_ALIGN_MIDDLE,
    DIYA_CAIRO_WIDGET_ALIGN_END,
    DIYA_CAIRO_WIDGET_ALIGN_MAX, // not used
    /*fill enum*/
    DIYA_CAIRO_WIDGET_FILL_NONE, // using extent size
    DIYA_CAIRO_WIDGET_FILL_CONTENT, // extent calculated by its content
    DIYA_CAIRO_WIDGET_FILL_EXTEND, // extend to the entire free space of the parent
    DIYA_CAIRO_WIDGET_FILL_MAX, //not used

    /* Visibility enum*/
    DIYA_CAIRO_WIDGET_VISIBLE_ON,
    DIYA_CAIRO_WIDGET_VISIBLE_OFF, // not show but take space
    DIYA_CAIRO_WIDGET_VISIBLE_HIDDEN, // not show and does not take space
    DIYA_CAIRO_WIDGET_VISIBLE_MAX, // not used

};

struct _DiyaCairoWidgetClass
{
    DiyaObjectClass parent_class;
    void (*draw)(gpointer, cairo_t*);
    void (*update)(gpointer);
    void (*resize)(gpointer);
    void (*align)(gpointer);
    void (*content_size)(gpointer, guint*, guint*);
    diya_rect_t (*alloc_space)(gpointer, gpointer);

    const gchar* name;
};

diya_point_t diya_cairo_widget_get_position(gpointer object);
diya_point_t diya_cairo_widget_get_global_position(gpointer object);
DiyaCairoWidget* diya_cairo_widget_get_parent(gpointer object);
diya_rect_t diya_cairo_widget_get_extent(gpointer object);
gboolean diya_cairo_widget_is_visible(gpointer object);
gboolean diya_cairo_widget_is_hidden(gpointer object);
gboolean diya_cairo_widget_is_damaged(gpointer object);

void diya_cairo_widget_damage(gpointer object);

void diya_cairo_widget_set_size(gpointer object, guint width, guint height);
void diya_cairo_widget_set_position(gpointer object, gint x, gint y);
void diya_cairo_widget_fill(gpointer object, diya_color_t * color);
void diya_cairo_widget_set_stroke_size(gpointer object, gdouble size);
void diya_cairo_widget_set_stroke_color(gpointer object, diya_color_t * color);

#define DIYA_TYPE_CAIRO_CONTAINER  (diya_cairo_container_get_type ())
G_DECLARE_DERIVABLE_TYPE (DiyaCairoContainer, diya_cairo_container, DIYA, CAIRO_CONTAINER, DiyaCairoWidget)


struct _DiyaCairoContainerClass
{
    DiyaCairoWidget parent_class;
};

void diya_cairo_container_add(gpointer object, gpointer child);

#endif