#include <assert.h>
#include "cairo-window.h"

/*
enum
{
    NO_PROP,
    BOX_MODE,
    N_PROPERTIES
};
static GParamSpec *g_prop[N_PROPERTIES] = {0};
*/
struct _DiyaCairoWindow
{
    DiyaCairoWidget * parent_object;
    DiyaCairoWidget * root;
};

G_DEFINE_FINAL_TYPE(DiyaCairoWindow, diya_cairo_window, DIYA_TYPE_CAIRO_WIDGET)

static void diya_cairo_window_dispose(GObject * object)
{
    g_debug("diya_cairo_window_dispose: %s", diya_object_to_string(object));
    DiyaCairoWindow * window = DIYA_CAIRO_WINDOW(object);
    if(window->root)
    {
        g_object_unref(window->root);
    }
    G_OBJECT_CLASS(diya_cairo_window_parent_class)->dispose(object);
}

static void diya_cairo_window_init(DiyaCairoWindow * window)
{
    window->root = NULL;
}

static void cairo_window_draw(gpointer object, cairo_t * cairo)
{
    DiyaCairoWindow * window = DIYA_CAIRO_WINDOW(object);
    DIYA_CAIRO_WIDGET_CLASS(diya_cairo_window_parent_class)->draw(object, cairo);
    if(window->root)
    {
        DIYA_CAIRO_WIDGET_GET_CLASS(DIYA_CAIRO_WIDGET(window->root))->draw(object, cairo);
    }
}

static void cairo_window_update(gpointer object)
{
    DiyaCairoWindow * self = DIYA_CAIRO_WINDOW(object);
    DIYA_CAIRO_WIDGET_CLASS(diya_cairo_window_parent_class)->update(object);

    if(self->root)
    {
        DIYA_CAIRO_WIDGET_CLASS(DIYA_CAIRO_WIDGET(self->root))->update(self->root);
    }
}

static void diya_cairo_window_class_init(DiyaCairoWindowClass* class)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS(class);
    DiyaCairoWidgetClass *wclass = DIYA_CAIRO_WIDGET_CLASS(class);
    wclass->draw = cairo_window_draw;
    wclass->update = cairo_window_update;
    DIYA_CAIRO_WIDGET_SET_NAME(class, "DiyaCairoWindow");
    gobject_class->dispose = diya_cairo_window_dispose;
}

void diya_cairo_window_set_root(gpointer object,  gpointer root)
{
    assert(DIYA_IS_CAIRO_WINDOW(object));
    assert(DIYA_IS_CAIRO_WIDGET(root));
    DiyaCairoWindow * window = DIYA_CAIRO_WINDOW(object);
    window->root = DIYA_CAIRO_WIDGET(root);
    diya_rect_t extent = diya_cairo_widget_get_extent(window);
    g_object_set(root, 
        "parent", window,
        "default-extent", &extent,
        "vspace", DIYA_CAIRO_WIDGET_FILL_EXTEND,
        "hspace", DIYA_CAIRO_WIDGET_FILL_EXTEND,
        NULL
        );
}

void diya_cairo_window_render(gpointer object, cairo_surface_t * surface)
{
    assert(DIYA_IS_CAIRO_WINDOW(object));
    DiyaCairoWidget * widget = DIYA_CAIRO_WIDGET(object);
    cairo_t *cr = cairo_create(surface);
    DIYA_CAIRO_WIDGET_GET_CLASS(widget)->draw(object,cr);
    cairo_destroy(cr);
}

void diya_cairo_window_update(gpointer object)
{
    assert(DIYA_IS_CAIRO_WINDOW(object));
    DIYA_CAIRO_WIDGET_CLASS(DIYA_CAIRO_WIDGET(object))->update(object);
}